# Visual UV
# Copyright (C) 2023 Samuel Bílek

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from . visualuv_ui import *
from . visualuv_ops import *

import bpy
from bpy.props import BoolProperty, FloatVectorProperty, FloatProperty, EnumProperty, PointerProperty


bl_info = {
    "name": "VisualUV",
    "author": "Samuel Bílek",
    "description": "Adds UV visualisation overlays for meshes. Useful for visualising UV stretching, individual UV islands and flipped UV normals.",
    "blender": (3, 4, 0),
    "version": (0, 0, 1),
    'location': 'VIEW_3D > Tools > VisualUV',
    "warning": "",
    "category": "Object"
}

classes = (
    VISUALUV_PT_uv_tool_menu3d,
    VISUALUV_PT_uv_tool_menu2d,
    VISUALUV_OT_update,
    VISUALUV_OT_toggle_texture,
    VISUALUV_OT_toggle_stretching,
    VISUALUV_OT_toggle_islands,
    VISUALUV_OT_toggle_normals,
    VISUALUV_OT_toggle_overlap,
    VISUALUV_OT_overlay,
)

def register():
    for c in classes:
        bpy.utils.register_class(c)

    bpy.types.WindowManager.uv_select_overlap = BoolProperty(default=False)
    
    bpy.types.Object.uv_enabled = BoolProperty()
    bpy.types.Object.uv_auto_update = BoolProperty(
        default=True, 
        description="Overlay auto-updates after key release or click. High count of geometry causes stutters"
    )
    bpy.types.Object.uv_backface_culling = BoolProperty(
        description="Turn on to cull backfaces"
    )
    bpy.types.Object.uv_show_wire = BoolProperty(
        default=True,
        description="Display wireframe in Edit-Mode, with seams, selected edges and vertices"
    )

    bpy.types.Object.uv_alpha = FloatProperty(
        default=0.75,
        max=1.0,
        min=0.0
    )

    bpy.types.Object.uv_recalculate = BoolProperty()
    bpy.types.Object.uv_overlap_recalculate = BoolProperty()
    bpy.types.Object.uv_show_3D = BoolProperty(
        default=True,
        description="Display overlay for the 3D Viewport"
    )
    bpy.types.Object.uv_show_2D = BoolProperty(
        default=True,
        description="Display overlay for the UV Editor"
    )
    bpy.types.Object.uv_enable_position_change = BoolProperty(
        description="Enable for position manipulation"
    )
    bpy.types.Object.uv_enable_color_change = BoolProperty(
        description="Enable for color manipulation"
    )
    bpy.types.Object.uv_enable_explosion_view = BoolProperty(
        description="Enable for separating and offsetting individual UV islands"
    )
    bpy.types.Object.uv_checker_texture = BoolProperty(
        default=False
    )
    bpy.types.Object.uv_fill_texture = BoolProperty(
        default=True,
        description="Fill the background with the selected texture of an active object"
    )
    bpy.types.Object.uv_image = PointerProperty(name='Image', type=bpy.types.Image)

    bpy.types.Object.uv_texture_scale = FloatProperty(
        default=1.0,
        soft_max=2.0,
        min=0.01
    )
    bpy.types.Object.uv_texture_alpha = FloatProperty(
        default=1.0,
        max=1.0, 
        min=0.0
    )
    bpy.types.Object.uv_max_difference = FloatProperty(
        default=0.25,
        min=0.01,
        soft_max=1.0
    )
    bpy.types.Object.uv_max_division = FloatProperty(
        default=2.0,
        min=1.01,
        soft_max=5.0
    )

    bpy.types.Object.uv_hue_shift = FloatProperty(
        default=0.0,
        max=1.0, 
        min=0.0
    )
    bpy.types.Object.uv_saturation = FloatProperty(
        default=1.0,
        max=1.0,
        min=0.0
    )
    bpy.types.Object.uv_value = FloatProperty(
        default=1.0,
        max=1.0, 
        min=0.0
    )

    bpy.types.Object.uv_explosion_offset = FloatProperty(
        default=0.0,
        soft_max=3.0,
        min=0.0)
    bpy.types.Object.uv_location_offset = FloatVectorProperty(
        subtype='XYZ'
    )
    bpy.types.Object.uv_operation = EnumProperty(
        items=[
            ('NONE', "None", "", 1),
            ('UV_STRETCHING', "Stretching", "", 2),
            ('UV_ISLANDS', "Islands", "", 3),
            ('UV_NORMALS', "Normals", "", 4),
            ('UV_OVERLAP', "Overlap", "", 5)
        ],
        default='NONE'
    )
    bpy.types.Object.uv_stretch_type = EnumProperty(
        items=[
            ('ANGLES', "Angles", "Difference between angles of UV and original Face", 1), 
            ('AREA', "Area", "Difference between the real and the average UV Face area", 2), 
            ('EDGE_LENGTH', "Edge length", "Difference between ratio of edge lengths of UV and original Face", 3)
        ], 
        default='ANGLES'
    )


def unregister():
    del bpy.types.Object.uv_stretch_type
    del bpy.types.Object.uv_operation
    del bpy.types.Object.uv_location_offset
    del bpy.types.Object.uv_explosion_offset
    del bpy.types.Object.uv_value
    del bpy.types.Object.uv_saturation
    del bpy.types.Object.uv_hue_shift
    del bpy.types.Object.uv_max_division
    del bpy.types.Object.uv_max_difference
    del bpy.types.Object.uv_texture_alpha
    del bpy.types.Object.uv_texture_scale
    del bpy.types.Object.uv_image
    del bpy.types.Object.uv_fill_texture
    del bpy.types.Object.uv_checker_texture
    del bpy.types.Object.uv_enable_explosion_view
    del bpy.types.Object.uv_enable_color_change
    del bpy.types.Object.uv_enable_position_change
    del bpy.types.Object.uv_show_2D
    del bpy.types.Object.uv_show_3D
    del bpy.types.Object.uv_overlap_recalculate
    del bpy.types.Object.uv_recalculate
    del bpy.types.Object.uv_alpha
    del bpy.types.Object.uv_show_wire
    del bpy.types.Object.uv_backface_culling
    del bpy.types.Object.uv_auto_update
    del bpy.types.Object.uv_enabled
    del bpy.types.WindowManager.uv_select_overlap

    for c in classes:
        bpy.utils.unregister_class(c)


if __name__ == "__main__":
    register()
